import argparse
import logging
import sys

from lussac.core import LussacData, LussacParams, LussacPipeline, LussacSpikeSorter


def parse_arguments(args: list | None) -> str:
	"""
	Parses the arguments given when launching Lussac.

	@param args: list
		All the arguments to parse.
	@return params_file: str
		Path to the file containing Lussac's parameters.
	"""

	parser = argparse.ArgumentParser()
	parser.add_argument("params_file", help="Path to the json file containing Lussac's parameters.")
	args = parser.parse_args(args)

	return args.params_file


def main() -> None:  # pragma: no cover
	"""
	The main function to execute Lussac.
	"""

	# STEP 0: Loading the parameters into the LussacData object.
	params_file = parse_arguments(sys.argv[1:])
	params = LussacParams.load_from_json_file(params_file)
	data = LussacData.create_from_params(params)

	# STEP 1: Running the spike sorting.
	if 'spike_sorting' in params:
		logging.info("Running spike-sorting algorithms:\n")
		for name, params0 in params['spike_sorting'].items():
			spike_sorter = LussacSpikeSorter(data.recording, name)
			data.sortings[name] = spike_sorter.launch(params0)

	# STEP 2: Running the pipeline.
	pipeline = LussacPipeline(data)
	pipeline.launch()


if __name__ == "__main__":  # pragma: no cover
	main()



import json
import pathlib
import platform

import jsmin

import lussac


class LussacParams:

	@staticmethod
	def load_from_string(params: str, params_folder: pathlib.Path | str | None = None):
		"""
		Loads the parameters from a string and returns them as a dict.

		@param params: str
			Lussac's parameters.
		@param params_folder: str
			Path to replace the "$PARAMS_FOLDER".
		"""

		if params_folder is not None:
			params_folder = str(pathlib.Path(params_folder).absolute())
			params = params.replace("$PARAMS_FOLDER", params_folder)
		if platform.system() == "Windows":  # pragma: no cover (OS specific).
			params = params.replace("\\", "\\\\")

		return json.loads(params)

	@staticmethod
	def load_from_json_file(filename: str, params_folder: pathlib.Path | str | None = None) -> dict:
		"""
		Loads the JSON parameters file and returns its content as a dict.

		@param filename: str
			Path to the file containing Lussac's parameters.
		@param params_folder: Path | str | None
			Path to replace the "$PARAMS_FOLDER".
			If None (default), will use the parent folder of the filename.
		@return params: dict
			Lussac's parameters.
		"""

		if params_folder is None:
			params_folder = str(pathlib.Path(filename).parent.absolute())
		else:
			params_folder = str(pathlib.Path(params_folder).absolute())

		with open(filename) as json_file:
			minified = jsmin.jsmin(json_file.read())  # Parses out comments.
			return LussacParams.load_from_string(minified, params_folder)

	@staticmethod
	def load_default_params(name: str, folder: pathlib.Path | str) -> dict:
		"""
		Loads the default parameters from the "params_example" folder.

		@param name: str
			The name of the default params file to load.
		@param folder: str
			Path to the folder where to create the "lussac" folder.
		@return params: dict
			The default parameters.
		"""

		if not name.startswith("params_"):
			name = f"params_{name}"
		if not name.endswith(".json"):
			name = f"{name}.json"

		params_folder = pathlib.Path(lussac.__file__).parent / "params_examples"
		file = params_folder / name

		return LussacParams.load_from_json_file(str(file), folder)



from dataclasses import dataclass
import logging
from pathlib import Path
import time

import lussac.utils as utils
import spikeinterface.core as si
import spikeinterface.curation as scur
import spikeinterface.preprocessing as spre
import spikeinterface.sorters as ss


@dataclass(slots=True)
class LussacSpikeSorter:
	"""
	Abstract class for all spike sorting algorithms running through Lussac.
	All children must have names to recognize them, and a launch method.

	Attributes:
		recording		A SpikeInterface recording object.
		name			The name of the analysis (not the sorting algorithm).
	"""

	recording: si.BaseRecording
	name: str

	def _preprocess(self, params: dict) -> None:
		"""
		Preprocesses the recording object based on what's in the params.

		@param params: dict
			Dictionary mapping the preprocessing functions to their arguments
		"""

		for preprocess_func, arguments in params.items():
			function = getattr(spre, preprocess_func)
			self.recording = function(self.recording, **arguments)

	def launch(self, params: dict) -> si.BaseSorting:
		"""
		Launches the spike sorting algorithm and returns the sorting object.

		@param params: dict
			The parameters for the spike sorting algorithm.
		"""

		sorter_name = params['sorter_name']
		folder = Path(params['sorter_params']['folder'])

		if (folder / "provenance.pkl").exists():
			sorting = si.load_extractor(folder / "provenance.pkl")
			logging.info(f"Loading analysis '{self.name}'\n")
			assert isinstance(sorting, si.BaseSorting)
			return sorting

		if 'preprocessing' in params:
			self._preprocess(params['preprocessing'])

		logging.info(f"Running spike-sorting for analysis '{self.name}' ")

		t_start = time.perf_counter()
		sorting = ss.run_sorter(sorter_name, self.recording, **params['sorter_params'])
		sorting = scur.remove_excess_spikes(sorting.remove_empty_units(), self.recording)
		sorting.annotate(name=self.name)
		sorting.dump_to_pickle(file_path=folder / "provenance.pkl", include_properties=True)
		t_end = time.perf_counter()

		logging.info(f"(done in {utils.format_elapsed_time(t_end - t_start)})\n")

		return sorting



from typing import Any
from overrides import override
import numpy as np
from lussac.core import MonoSortingModule
import lussac.utils as utils
import spikeinterface.core as si


class RemoveBadUnits(MonoSortingModule):
	"""
	Removes the bad units from the sorting object.
	The definition of "bad" unit is given by the parameters dictionary.
	"""

	@property
	@override
	def default_params(self) -> dict[str, Any]:
		return {
			'wvf_extraction': {
				'ms_before': 1.0,
				'ms_after': 1.0,
				'max_spikes_per_unit': 500,
				'filter_band': [150, 9000]
			}
		}

	@override
	def run(self, params: dict[str, Any]) -> si.BaseSorting:
		wvf_extraction_params = params.pop('wvf_extraction', {})
		if self.analyzer is None:
			self.precompute_analyzer(params)

		units_to_remove = np.zeros(self.sorting.get_num_units(), dtype=bool)
		reasons_for_removal = np.array([''] * self.sorting.get_num_units(), dtype=object)

		for attribute, p in params.items():
			if attribute == "all":
				units_to_remove[:] = True
				reasons_for_removal[:] += " ; all"
				break

			value = self.get_units_attribute_arr(attribute, p, **wvf_extraction_params)
			if 'min' in p:
				units_to_remove |= value < p['min']
				reasons_for_removal[value < p['min']] += f" ; {attribute} < {p['min']}"
			if 'max' in p:
				units_to_remove |= value > p['max']
				reasons_for_removal[value > p['max']] += f" ; {attribute} > {p['max']}"

		sorting = self.sorting.select_units([unit_id for unit_id, bad in zip(self.sorting.unit_ids, units_to_remove) if not bad])
		bad_sorting = self.sorting.select_units([unit_id for unit_id, bad in zip(self.sorting.unit_ids, units_to_remove) if bad])
		reasons_for_removal = ["Reason(s) for removal: " + reason[3:] for reason in reasons_for_removal[units_to_remove]]

		self._plot_bad_units(bad_sorting.unit_ids, reasons_for_removal)

		return sorting

	def precompute_analyzer(self, params: dict[str, Any]) -> None:
		params = self.update_params(params)
		wvf_extraction = params['wvf_extraction']

		self.create_analyzer(filter_band=wvf_extraction['filter_band'], cache_recording=True)
		self.analyzer.compute({
			'random_spikes': {'max_spikes_per_unit': wvf_extraction['max_spikes_per_unit']},
			'templates': {'ms_before': wvf_extraction['ms_before'], 'ms_after': wvf_extraction['ms_after']},
			'spike_amplitudes': {'peak_sign': 'both'}
		})

	def _plot_bad_units(self, bad_unit_ids, reasons_for_removal: list[str]) -> None:
		"""
		Plots the units that were removed.

		@param bad_unit_ids:
			The unit ids that were removed.
		"""
		if len(bad_unit_ids) == 0:
			return

		analyzer = self.analyzer.select_units(bad_unit_ids, format="memory")

		annotations = [{'text': reason, 'x': 0.6, 'y': 1.02, 'xref': "paper", 'yref': "paper", 'xanchor': "center", 'yanchor': "bottom", 'showarrow': False} for reason in reasons_for_removal]
		utils.plot_units(analyzer, filepath=f"{self.logs_folder}/bad_units", annotations_change=annotations)
